@@ -31,12 +31,12 @@ def print_metadata_tables_to_file(dw, study_id):
     file_dir = "reports/"
     timestamp = datetime.datetime.now()  # use the current date and time if none is specified
     time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
-    fname = mk_txt_report_file_name(file_dir, "metadata-table", time_fname_str)
+    fname = mk_txt_report_file_name(file_dir, "metadata-table-", time_fname_str)
 
     metadata = create_measurement_group_info(dw, study_id)
     headers = ['Measurement Type', 'Id', 'Value Type', 'Optional?', 'Units', 'Min Value', 'Max Value', 'Categories']
 
-    with open(fname, "w") as f:
+    with open(fname, "w", encoding="utf-8") as f:
         print(f'Measurement Groups for Study {study_id}', file=f)
         for mg_id, mg_data in metadata.items():
             mg_info = mg_data['message_types']
